#ifndef LIGHT_SOURCE_H
#define LIGHT_SOURCE_H

// In this file we have classes that define some simple light sources for
// raytracing. There should be one Lighting object in a scene, which contains
// the color of ambient illumination along with an array of directed light sources.
// The DirectedLightSource class is a pure virtual base class for lights which have
// direction associated with them: the subclasses here are infinitely distant lights
// (where the directions of light rays are always parallel) and point lights (where
// the rays emanate from a single point, and fall off in intensity with an inverse
// square law).
//
// Note throughout that Vec3f (a vector of 3 floats) is used for RGB colors, which
// is thought of as a measure of the intensity of light.

#include <cfloat>
#include <vector>
#include "primitives.h"

struct DirectedLight
{
   virtual ~DirectedLight(void) {}

   // Given a point x in the scene, return in the other variables the following information:
   //  light_direction: a unit-length vector pointing from x towards the light
   //  light_color: the RGB intensity of the light at x
   //  light_distance: how far away the light is from x
   virtual void evaluate(const Vec3f& x, Vec3f& light_direction, Vec3f& light_color, float& light_distance) const = 0;
};

struct DistantLight: public DirectedLight
{
   Vec3f direction; // the direction light is travelling in; should be unit-length
   Vec3f color; // its RGB intensity

   // the default is an overhead white light
   DistantLight(void) : direction(0,-1,0), color(1,1,1) {}

   DistantLight(const Vec3f& direction_, const Vec3f& color_) : direction(normalized(direction_)), color(color_) {}

   virtual void evaluate(const Vec3f& x, Vec3f& light_direction, Vec3f& light_color, float& light_distance) const
   {
      light_direction=-direction; // note the negative: light_direction should point towards the light
      light_color=color;
      light_distance=FLT_MAX; // this isn't quite infinity, but is the largest finite float
   }
};

struct PointLight: public DirectedLight
{
   Vec3f position; // the source of the light
   Vec3f color;

   // the default is a white light at the origin
   PointLight(void) : position(0,0,0), color(1,1,1) {}

   PointLight(const Vec3f& position_, const Vec3f& color_) : position(position_), color(color_) {}

   virtual void evaluate(const Vec3f& x, Vec3f& light_direction, Vec3f& light_color, float& light_distance) const
   {
      light_direction=position-x; // un-normalized: we'll have to fix it later to be unit-length
      float distance_squared=mag2(light_direction)+1e-30; // bias it to be strictly positive
      light_distance=std::sqrt(distance_squared);
      light_direction/=light_distance; // normalize the direction to unit-length
      light_color=color/distance_squared; // fade out the light according to the inverse square law
   }
};

struct Lighting
{
   Vec3f ambient;
   std::vector<DirectedLight*> directed_light;

   Lighting(void) : ambient(0,0,0) {}
};

#endif
