#include <iostream>
#include "camera.h"
#include "image_io.h"
#include "mesh_io.h"

int main(void)
{
   // What follows is an example scene, a plane with ten reflective spheres in a ring, and a
   // red triangle mesh floating overhead (loaded in from the file "example.obj").
   // To change the scene, you have to change this code and recompile at the moment.

   // set up lights
   Lighting lighting;
   lighting.ambient=Vec3f(0.1,0.1,0.1);
   lighting.directed_light.push_back(new DistantLight(Vec3f(-1, -4, 0), 0.2f*Vec3f(0.4,0.5,0.7)));
   lighting.directed_light.push_back(new PointLight(Vec3f(0.5, 4, 1), 10*Vec3f(1,0.8,0.6)));

   // set up shapes
   AggregateShape world;
   for(int i=0; i<10; ++i){
      Sphere *sphere=new Sphere(Vec3f(std::cos(2*M_PI*0.1*i), 0.5, std::sin(2*M_PI*0.1*i)), 0.25);
      sphere->material.diffuse_color=Vec3f(0.2,0.2,0.2);
      sphere->material.mirror_color=Vec3f(0.8,0.8,0.8);
      world.subshape.push_back(sphere);
   }
   {
      TriangleMesh *mesh=new TriangleMesh;
      read_objfile(mesh->tri, mesh->x, mesh->normal, "example.obj");
      // before adding this triangle mesh to the scene, we'll transform it a bit (scale, rotate, translate)
      Transformation M=translate(Vec3f(0,0.9,0))*rotate_y(70)*rotate_z(20)*rotate_x(40)*scale(1.2,1.2,1.2);
      // transform all the points in the mesh
      for(unsigned int i=0; i<mesh->x.size(); ++i) mesh->x[i]=M.apply_to_point(mesh->x[i]);
      if(!mesh->normal.empty()){ // if vertex normals were specified in the mesh
         // we also have to transform the normals, this time using the inverse transpose of the above matrix
         Transformation N=transpose(inverse(M));
         for(unsigned int i=0; i<mesh->normal.size(); ++i) mesh->normal[i]=normalized(N.apply_to_direction(mesh->normal[i]));
      }
      mesh->build_tree();
      mesh->material.diffuse_color=Vec3f(1,0.3,0.2);
      mesh->material.mirror_color=Vec3f(0.05,0.05,0.05);
      world.subshape.push_back(mesh);
   }
   {
      Plane *plane=new Plane;
      world.subshape.push_back(plane);
   }
   world.build_tree();

   // set up raytracer
   // (this has to come after lighting and world, since raytracer needs a reference to these objects)
   RayTracer raytracer(lighting, world);

   // and finally set up the camera
   // (this has to come after raytracer, since the camera needs a reference to it)
   Camera camera(raytracer);
   camera.modelview=translate(Vec3f(0,0,-4))*rotate_x(45);
   camera.left=-0.5;
   camera.right=0.5;
   camera.bottom=-0.375;
   camera.top=0.375;

   // do the render
   Array2<Vec3f> image;
   camera.render_image(image);

   // and save the output image
   write_sgi(image, "output.sgi");

   return 0;
}

